module Interpreter where

import Prelude hiding (lookup,update,enter)
import Data.Maybe(fromJust, listToMaybe, fromMaybe)
import Data.Char(isDigit, toLower)

import qualified Data.Map as M

import AST
import State

-- Elementary functions on values

-- Representing booleans
isTrue :: Val-> Bool
isTrue (N d) = d /= 0

boolVal :: Bool-> Val
boolVal b = N $ if b then 1 else 0


--
-- Executing a single command
--
exec :: Cmd -> St ()
exec w@(While e cs) = do
   v<- evaluate e
   vs <- get vars
   if isTrue v then do {execs cs; exec w} else return ()
exec (If e cs1 cs2) = do
   v<- evaluate e
   if isTrue v then execs cs1 else execs cs2
exec (Assign i e) = do
   v<- evaluate e
   set $ \s-> s{vars= M.insert i v (vars s)}
exec (Print e) = do
   v<- evaluate e
   set $ \s-> s{output= show v : output s}

-- Executing a sequence of commands
execs :: [Cmd]-> St ()
execs = mapM_ exec

-- Evaluating an expression, errors propagated into St
evaluate :: Expr-> St Val
evaluate e = do
   ev<- get $ eval e
   case ev of Right v -> return v
              Left  e -> err e

-- Evaluating an expression: has no return state, only errors
eval :: Expr-> State-> Error Val
eval (Var i) s =
  case M.lookup i (vars s) of
    Just v-> return v
    Nothing->  Left $ "Undeclared variable: "++ i
eval (Number d) _ = return $ N d
eval (Negate e) s = do
  v <- eval e s
  case v of
    N d -> return $ N (-d)
eval (Plus e1 e2) s = do
  v1 <- eval e1 s
  v2 <- eval e2 s
  case (v1, v2) of
    (N d1, N d2) -> return $ N (d1+ d2)
eval (Times e1 e2) s = do
  v1 <- eval e1 s
  v2 <- eval e2 s
  case (v1, v2) of
    (N d1, N d2) -> return $ N $ d1* d2
eval (Div e1 e2) s = do
  v1 <- eval e1 s
  v2 <- eval e2 s
  case (v1, v2) of
    (N d1, N d2) | d2 == 0 -> Left "Division by zero."
    (N d1, N d2) -> return $ N $ d1/d2
eval (LessEq e1 e2) s = do
  v1 <- eval e1 s
  v2 <- eval e2 s
  case (v1, v2) of
    (N d1, N d2) -> return $ boolVal (d1 <= d2)
eval (Eq e1 e2) s = do
  v1 <- eval e1 s
  v2 <- eval e2 s
  case (v1, v2) of
    (N d1, N d2) -> return $ boolVal (d1 == d2)

-- Evaluating a sequence of declarations
decls :: [Id]-> St ()
decls = mapM_ (\i-> set $ \s-> s{vars= M.insert i (N 0) (vars s)})

-- Evaluating a whole program
prog :: Prog-> IO ()
prog (Prog ds cs) = do
  let i = initialState 
  case run (do {decls ds; execs cs}) i of
     Left err -> putStrLn $ "*** ERROR: "++ err
     Right ((), s) -> do
        mapM_ putStrLn (reverse (output s))

initialState :: State
initialState =
  State { vars = M.empty
        , output = []
        }
